import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter_vasat/models/session.dart';
import 'package:flutter_vasat/models/vasat_model.dart';
import 'package:flutter_vasat/storage/storage_provider.dart';
import 'package:flutter_vasat/vasat_error.dart';
import 'package:path_provider/path_provider.dart';
import 'package:sembast/sembast.dart';
import 'package:sembast/sembast_io.dart';
import 'package:sembast_web/sembast_web.dart';
import 'package:path/path.dart';
import '../vasat_types.dart';

class VasatSembastStorageProvider extends VasatStorageProvider {

  Database? db;

  VasatSembastStorageProvider();

  @override
  Future<void> openDatabase(String dbName) async {

    // Check if platform is web or mobile in order to use the right underlining library and DB. Web uses indexedDB. Mobile uses single file in local storage
    if (kIsWeb) {
      // Open the database
      db = await databaseFactoryWeb.openDatabase(dbName);
    } else {
      // get the application documents directory
      var dir = await getApplicationDocumentsDirectory();
      // make sure it exists
      await dir.create(recursive: true);
      // build the database path
      var dbPath = join(dir.path, '$dbName.db');
      // open the database
      db = await databaseFactoryIo.openDatabase(dbPath);
    }
  }

  @override
  Future<int?> getObjectLastTimestamp(String vasatObjectName) => getSimpleObject<int>(vasatObjectName);

  @override
  Future<void> setObjectLastTimestamp(String vasatObjectName, int timestamp) => setSimpleObject(vasatObjectName, timestamp);

  Future<T?> getSimpleObject<T extends Object>(String objectName, {VasatFromJsonFactory<T>? factory} ) {
    if (db != null) {
      var store = StoreRef.main();
      return store.record( getObjectKey(objectName) ).get(db!).then((value) {
        if(value != null) {
          return (factory != null) ? factory(value) : value as T;
        } else {
          return null;
        }
      });
    } else {
      throw VasatError("DB is null or has not been initializes. Make sure openDatabase() has been called ", status: 5100);
    }

  }

  Future<bool> setSimpleObject(String objectName, Object object) {
    debugPrint("ENTER setSimpleObject $objectName");
    if (db != null) {
      var store = StoreRef.main();
      debugPrint("ENTER setSimpleObject $object");
      return store.record( getObjectKey(objectName) ).put(db!, object).then((value) => value != null);
    } else {
      throw VasatError("DB is null or has not been initializes. Make sure openDatabase() has been called ", status: 5100);
    }
  }

  Future<bool> removeSimpleObject(String objectName) {
    if (db != null) {
      var store = StoreRef.main();
      return store.record( getObjectKey(objectName) ).delete(db!).then((value) => value != null);
    } else {
      throw VasatError("DB is null or has not been initializes. Make sure openDatabase() has been called ", status: 5100);
    }
  }

  @override
  Future<bool> saveObjects<T extends VasatModel<T>>(List<T> objects) {
    if (db != null) {
      // First get the vasat object name
      var objectName = objects.first.getObjectName();
      var store = stringMapStoreFactory.store(objectName);
      var keys = objects.map((e) => e.id.toString()).toList();
      var values = objects.map((e) => e.toJson()).toList();
      return store.records(keys).put(db!, values, merge: true).then((value) => true);
    } else {
      throw VasatError("DB is null or has not been initializes. Make sure openDatabase() has been called ", status: 5100);
    }
  }

  @override
  Future<List<T>> getObjects<T extends VasatModel<T>>(VasatFromJsonFactory<T> factory) {
    if (db != null) {
      // First get the vasat object name from the factory
      var objectName = factory({}).getObjectName();
      var store = stringMapStoreFactory.store(objectName);
      return store.find(db!).then((res) => res.map((record) => factory(record.value) ).toList() );
    } else {
      throw VasatError("DB is null or has not been initializes. Make sure openDatabase() has been called ", status: 5100);
    }
  }

  @override
  Future<Session?> getSession() => getSimpleObject<Session>("Session", factory: (data) => Session.fromJson(data));

  @override
  Future<bool> removeSession() => removeSimpleObject("Session");

  @override
  Future<bool> saveSession(Session session) => setSimpleObject("Session", session.toJson());



}