import 'dart:convert';
import 'package:flutter_vasat/models/session.dart';
import 'package:flutter_vasat/models/vasat_model.dart';
import 'package:flutter_vasat/storage/storage_provider.dart';
import 'package:shared_preferences/shared_preferences.dart';
import '../vasat_types.dart';

class VasatSharedPreferencesProvider extends VasatStorageProvider {

  VasatSharedPreferencesProvider();

  @override
  Future<void> openDatabase(String dbName) {
    this.dbName = dbName;
    return new Future<bool>.value(true);
  }

  @override
  Future<int?> getObjectLastTimestamp(String vasatObjectName) {
    return SharedPreferences.getInstance()
      .then((prefs) {
        return prefs.getInt(this.dbName + "_" + vasatObjectName);
      });
  }

  @override
  Future<void> setObjectLastTimestamp(String vasatObjectName, int timestamp) {
    return SharedPreferences.getInstance()
      .then((prefs) {
        return prefs.setInt(this.dbName + "_" + vasatObjectName, timestamp);
      });
  }

  @override
  Future<bool> saveObjects<T extends VasatModel<T>>(List<T> objects) {
    return SharedPreferences.getInstance().then((prefs) {

      // Get first object to determine the vasat object name
      String objName = objects.first.getObjectName();
      // Get already stored object list as a Map<String,Object> where key is the id
      var objs = prefs.getString(objName);
      Map<String,dynamic> objsMap = {};
      if (objs != null) objsMap = jsonDecode(objs);
      // generate new list of Map objects
      Map<String,dynamic> newObjsMap = {};
      objects.forEach((element) {
        String uid = (element.id == null) ? generateUID(T) : element.id!;
        if (element.id == null) element.id = uid;
        newObjsMap[uid] = element.toJson();
      } );

      // Concatenate both maps replacing values on old Map with new Map if keys/id are already present.
      objsMap.addAll(newObjsMap);

      return prefs.setString( objName , jsonEncode(objsMap));
    });
  }

  @override
  Future<List<T>> getObjects<T extends VasatModel<T>>(VasatFromJsonFactory<T> factory) {
    // use factory to construct empty object just to get the vasat object name.
    var vasatObjectName = factory({}).getObjectName();
    return SharedPreferences.getInstance().then((prefs) {
      var stringObjs = prefs.getString( vasatObjectName );
      Map<String, dynamic> mapObjs = {};
      if (stringObjs != null) mapObjs = jsonDecode(stringObjs);
      List<T> objList = mapObjs.entries.map((ele) => factory(ele.value) ).toList();
      return objList;
    });
  }

  @override
  Future<Session?> getSession() {
    return SharedPreferences.getInstance().then((prefs) {
      var sesString = prefs.getString( getObjectKey("Session") );
      Session? session;
      if (sesString != null) {
        session = Session(json: jsonDecode(sesString));
      }
      return session;
    });
  }

  @override
  Future<bool> removeSession() {
    return SharedPreferences.getInstance().then((prefs) => prefs.remove( getObjectKey("Session") ) );
  }

  @override
  Future<bool> saveSession(Session session) {
    return SharedPreferences.getInstance().then((prefs) => prefs.setString( getObjectKey("Session"), jsonEncode(session.toJson())));
  }
}