library vasat;

import 'dart:convert';
import 'dart:io';
import 'package:flutter/cupertino.dart';
import 'package:flutter_vasat/models/device.dart';
import 'package:flutter_vasat/models/user.dart';
import 'package:flutter_vasat/storage/shared_preferences_provider.dart';
import 'package:flutter_vasat/storage/storage_provider.dart';
import 'package:flutter_vasat/vasat_error.dart';
import 'package:http/http.dart' as http;
import 'package:path_provider/path_provider.dart';
import 'package:rxdart/subjects.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'dart:async';
import 'models/vasat_model.dart';
import 'vasat_config.dart';
import 'models/session.dart';
import 'vasat_types.dart';

/// Vasat flutter SDK
class Vasat {

  final String name;
  final VasatConfig config;
  VasatStorageProvider storageProvider;
  Session? _session;
  Session? get currentSession => _session;

  final _sessionController = BehaviorSubject<Session?>();
  Stream<Session?> get sessionStream => _sessionController;


  static const String defaultName = "vasat";
  // _cache is library-private, thanks to the _ in front of its name.
  static final Map<String, Vasat> _cache = <String, Vasat>{};


  Vasat({this.name = defaultName, required this.config, required this.storageProvider }) {
    // Replace it in case we want to change config at any stage programmatically.
    _cache[name] = this;
  }

  /// Additional method that does the same as the factory constructor. Just here for making it more natural to get instances of Vasat
  static Vasat instance([String name = defaultName]) {
    // Return cached vasat for given name. If not found in cache return a new empty Vasat instace
    return _cache[name] ?? Vasat(config: VasatConfig(clientId: '', host: ''), storageProvider: VasatSharedPreferencesProvider());
  }

  T getStorageProvider<T extends VasatStorageProvider>() {
    return storageProvider as T;
  }

  /// Initializes the SDK local storage and check if there is a surrent session.
  Future<Session?> initialize() {
    return storageProvider.openDatabase(name).then((value) => loadCurrentSession());
  }

  /// Call this to dispose streams. Should be called when the app is closed
  void dispose() {
    _sessionController.close();
  }

  //////////////////////////////////////////////Authentication/////////////////////////////////////////////////////////
  Future<Session> loginWithClientCredentials() async {
    Device device = await Device.mkDevice();
    var body = <String, dynamic>{
      "grant_type": "client_credentials",
      "client_id": this.config.clientId,
      "client_secret": this.config.clientSecret,
      "device": device.toJson()
    };
    return apiPost("/access/token", body)
        .then((res) => processAccessToken(res))
        .then((session) =>  session );
  }

  Future<Session> loginWithUserCredentials(String username, String password) async {
    var body = <String, String>{
      "grant_type": "password",
      "client_id": this.config.clientId,
      "username": username,
      "password": password,
    };
    return apiPost("/access/token", body)
        .then((res) => processAccessToken(res));
    // TODO UNCOMMENT when server side fixed.
    //.then((session) => apiMe().then((u) => session) )
    //.then((session) =>  createDevice().then((value) => session) );
  }

  Future<Session> signUpUser(String username, String password, {String? name, String? email}) async {
    var props = <String, dynamic>{
      "username": username,
      "password": password
    };
    if (name != null) props["name"] = name;
    if (email != null) props["email"] = email;
    return signUpUserWithProps(props);
  }

  Future<Session> signUpUserWithProps(Map<String, dynamic> properties) async {
    var body = <String, dynamic>{
      "grant_type": "password",
      "client_id": this.config.clientId,
    };
    if (properties != null) body.addAll(properties);
    return apiPost("/access/signup", body)
        .then((res) => processAccessToken(res)
        .then((session) => apiMe().then((u) => session)) );
  }

  Future<void> logout() async {
    return apiGet("/access/logout").then((res) {
      _session = null;
      // Notify session has been terminated/cancelled/deleted
      _sessionController.add(null);
      storageProvider.removeSession();
    });
  }

  Future<dynamic> passwordReset(String email) async {
    var body = <String, String>{
      "email": email,
    };
    return apiPost("/access/forget", body);
  }

  Future<Session> processAccessToken(Map<String, dynamic> response) async {
    debug("processSession: Processing access token response: $response");
    _session = Session.fromJson(response);
    // Notify session listeners we have a new session.
    _sessionController.add(_session);
    debug("processSession: Access token parsed");
    return storageProvider.saveSession(_session!).then((value) => _session!);
  }
  /////////////////////////////////////////Session and Current User////////////////////////////////////////////////////

  Future<bool> hasActiveSession() async {
    if (_session == null)
      _session = await loadCurrentSession();

    return _isSessionValid(_session);
  }

  Future<Session?> loadCurrentSession() {
    debug("loading current session from storage");
    return storageProvider.getSession().then((session) {
      // When loading current session initialize _session attribute
      _session = null;
      // Check if it is valid, otherwise invalidate/remove it
      if (session != null && _isSessionValid(session)) {
        _session = session;
      } else if (session != null) {
        // If we had a session stored locally but is invalid, remove it
        storageProvider.removeSession();
      }
      // Trigger session stream
      _sessionController.add(_session);
      return _session;
    });
  }

  /// Validates if the session is valid.
  bool _isSessionValid(Session? session) {
    return (session != null && session.expiry == null);
  }

  Future<User> apiMe() async {
    return apiGet("/api/me").then((response) {
      debug("Success apiMe response ${response.toString()}");
      User user = User.fromJson(response);
      if (_session != null) {
        _session!.user = user;
        storageProvider.saveSession(_session!);
      }
      return user;
    });
  }

  ///////////////////////////////////////// CRUD ////////////////////////////////////////////////////

  Future<String> saveRemote<T extends VasatModel<T>>(T entity) async {
    
    return apiPost(getUrlPath<T>(entity), entity.toJsonItem())
        .then((res) => (res is Map && res.containsKey("message")) ? res["message"].toString() : res.toString() );

  }


  ///////////////////////////////////////// BaseModel end point base URL helpers ////////////////////////////////////////////////////
  String getObjectName<T extends Object>({String? vasatObjectName}) {
    return vasatObjectName ?? T.toString();
  }

  /*
  String getUrlRoot<T extends Object>({String? vasatObjectName}) {
    var url = "/api/${getObjectName<T>(vasatObjectName: vasatObjectName)}";
    //debug("getUrlRoot: $url");
    return url;
  }

   */

  String getUrlRoot(String vasatObjectName) {
    var url = "/api/$vasatObjectName";
    //debug("getUrlRoot: $url");
    return url;
  }


  String getUrlPath<T extends VasatModel<T>>(T entity) {
    var path = "";
    if (entity.id != null)
      path = "/${entity.id}";

    return getUrlRoot(entity.getObjectName()) + path;
  }

  String getObjectUrl<T extends VasatModel<T>>(T entity) {

    String uri;
    if (this.config != null && this.config.host != null) {
      uri = this.config.host + getUrlPath<T>(entity);
    } else {
      uri = "";
    }
    return uri;
  }

  ///////////////////////////////////////// CRUD FILES ////////////////////////////////////////////////////

  String getObjectImageUrl<T extends VasatModel<T>>(T entity) {
    return getObjectFileUrl<T>(entity, "image.png");
  }

  String getObjectFileUrl<T extends VasatModel<T>>(T entity, String fileName) {
    return getObjectUrl<T>(entity) + "/$fileName?access_token=${(_session != null) ? _session!.token : ""}";
  }

  Future<dynamic> getFilesInfo(String objectType, dynamic objectId) {
    return apiGet("/api/$objectType/$objectId.files");
  }

  Future<File> getFile(String objectType, dynamic objectId, String fileName, {String? saveToDirectory}) async {
    return await apiGetWithTokenUrl("/api/$objectType/$objectId/$fileName").then((fileRes) async {
      debug("Processing file:");
      File file;
      if (saveToDirectory != null) {
        file = File(saveToDirectory + "/$fileName");
      } else {
        var dir = await getTemporaryDirectory();
        var itemDir = Directory("${dir.path}/vasat/$objectType");
        if (!itemDir.existsSync()) {
          // Copy file to app directory location
          itemDir.createSync();
        }
        file = File("${itemDir.path}/$fileName");
      }
      file.writeAsBytesSync(fileRes, flush: true);
      debug("Processing file written");
      return file;
    });
  }

  Future<void> sendFiles(String objectType, dynamic objectId, String fileName, List<String> filePaths) {
    return apiMultipartPost("/api/$objectType/$objectId/$fileName", null, filePaths).then((value) {
      debug("ENTER sendFile success multipart");
    }).catchError((err) {
      debug("ENTER sendFile err ${err.toString()} multipart");
    });
  }

  Future<void> sendFile(String objectType, dynamic objectId, String fileName, File file) {
    return apiPostFile("/api/$objectType/$objectId/$fileName", file);
  }

  Future<dynamic> deleteFile(String objectType, dynamic objectId, String fileName) {
    return apiDelete("/api/$objectType/$objectId/$fileName");
  }

  ///////////////////////////////////////// Base API calls and methods ////////////////////////////////////////////////////

  Future<dynamic> apiDelete(String url) async {
    if (this.config != null && this.config.host != null) {
      var uri = Uri.parse(this.config.host + url);

      debug("apiDelete URL: ${uri.toString()}");

      return http.delete(uri, headers: getApiHeaders())
          .then( (res) => processResponse(res) );
    } else  {
      return Future.error(VasatError("No vasat configuration or host found", status: 8000));
    }

  }

  Future<dynamic> apiGet(String url, {Map<String, String>? params, bool returnPayloadOnly = true}) async {
    if (this.config != null && this.config.host != null) {
      // Maybe change this and uncomment below after Vasat2 search fixed and make sure behaves the same
      String queryParams = (params != null) ? params.entries.map((ele) => ele.key + "=" + ele.value).toList().join("&") : "";
      var uri = Uri.parse(this.config.host + url + ((queryParams.isNotEmpty) ? "?$queryParams" : "" ) );

      //var uri = Uri.parse(this.config.host + url);
      //if (params != null) uri = uri.replace(queryParameters: params);

      debug("apiGet URL: ${uri.toString()} With params: ${uri.queryParameters.toString()}");

      return http.get(uri, headers: getApiHeaders()).then((response) =>  processResponse(response, returnPayloadOnly));
    } else  {
      return Future.error(VasatError("No vasat configuration or host found", status: 8000));
    }

  }

  Future<dynamic> apiGetWithTokenUrl(String url, {Map<String, String>? params, bool returnPayloadOnly = true}) async {
    if (this.config != null && this.config.host != null) {
      if (_session != null && _session!.token != null) {
        if (params == null) params = Map();
        params["access_token"] = _session!.token!;
      }

      var uri = Uri.parse(this.config.host + url);
      if (params != null) uri = uri.replace(queryParameters: params);

      debug("apiGet URL: ${uri.toString()} With params: ${uri.queryParameters.toString()}");

      return http.get(uri).then((res) => processResponse(res, returnPayloadOnly));
    } else  {
      return Future.error(VasatError("No vasat configuration or host found",status: 8000));
    }

  }

  Future<dynamic> apiPut(String url, Map<String, dynamic> payload, {bool returnPayloadOnly = true}) async {
    if (this.config != null && this.config.host != null) {
      debug("apiPut URL: ${url} With body: ${jsonEncode(payload)}");
      return http.put( Uri.parse(this.config.host + url), body: jsonEncode(payload), headers: getApiHeaders(isJson: true))
          .then((res) => processResponse(res, returnPayloadOnly));
    } else  {
      return Future.error(VasatError("No vasat configuration or host found",status: 8000));
    }

  }

  Future<dynamic> apiPost(String url, Map<String, dynamic> payload, {bool returnPayloadOnly = true}) async {
    if (this.config != null && this.config.host != null) {
      debug("apiPost URL: ${this.config.host}${url} With body: ${jsonEncode(payload)}");
      return http.post( Uri.parse(this.config.host + url), body: jsonEncode(payload), headers: getApiHeaders(isJson: true))
          .then((res) => processResponse(res, returnPayloadOnly));
    } else  {
      return Future.error(VasatError("No vasat configuration or host found",status: 8000));
    }
  }

  Future<dynamic> apiPostFile(String url, File file) async {
    if (this.config != null && this.config.host != null) {
      debug("apiPostFile URL: ${url}");
      var headers = getApiHeaders(isJson: false);
      headers["Content-Type"] = "application/octet-stream";
      return http.post( Uri.parse(this.config.host + url), body: file.readAsBytesSync(), headers: headers);
    } else  {
      return Future.error(VasatError("No vasat configuration or host found",status: 8000));
    }
  }

  Future<dynamic> apiMultipartPost(String url, dynamic body, List<String> filePaths) async {
    if (this.config != null && this.config.host != null) {
      var request = http.MultipartRequest("POST", Uri.parse(this.config.host + url));
      String payload;
      if (body != null) {
        if (body is String)
          payload = body;
        else if (body is Map)
          payload = jsonEncode(body);
        else
          payload = jsonEncode(body);

        // Set payload fields
        request.fields["payload"] = payload;
      }

      for (var path in filePaths) {
        request.files.add(await getMultipartFile(path));
      }

      var res = await request.send();
      debug("Multipart response status: ${res.statusCode} stream: ${res.stream.toString()}");
    } else  {
      return Future.error("No vasat configuration or host found");
    }

  }

  Future<http.MultipartFile> getMultipartFile(String path) async {
    // infer content type
    String contentType;
    String fileName = path.substring(path.lastIndexOf("/")+1).toLowerCase();
    String ext = path.substring(path.lastIndexOf(".")+1).toLowerCase();
    switch(ext) {
      case "png":
        contentType = "application/octet-stream";
        break;
      case "jpg":
      case "jpeg":
        contentType = "application/octet-stream";
        break;
      default:
        contentType = "application/octet-stream";
        break;
    }
    //return http.MultipartFile.fromPath(fileName, path, contentType: MediaType.parse(contentType) );
    return http.MultipartFile.fromPath(fileName, path);
  }

  Map<String, String> getApiHeaders({bool isJson = false}) {
    //debug("getApiHeaders token: ${_session?.token}");
    var headers = <String,String>{};
    // If POST or PUT methods this should be set to true
    if (isJson) headers["Content-Type"] = "application/json";
    // Check if we have token
    if (_session != null && _session!.token != null) headers["Authorization"] = "Bearer " + _session!.token!;

    return headers;
  }

  Future<dynamic> processResponse(http.Response response, [bool returnPayloadOnly = true]) {
    debug("processResponse http status ${response.statusCode}");
    //debug("processResponse http headers ${response.headers}");
    //debug("processResponse http reasonPhrase ${response.reasonPhrase}");
    //debug("processResponse http res ${response.body}");
    // check response herader for the content_type and process accordingly.
    String? contentType = response.headers["content-type"];
    debug("processResponse contentType $contentType");
    if (contentType != null && contentType == "application/octet-stream") {
      return Future<dynamic>.value(response.bodyBytes);
    } else {
      Map<String, dynamic> resBody = jsonDecode(response.body);
      if (response.statusCode < 400 && resBody.containsKey("access_token") ) {
        return Future<dynamic>.value(resBody);
      } else if (response.statusCode < 400 && resBody.containsKey("payload") && resBody.containsKey("status") && (resBody["status"] < 400) && returnPayloadOnly) {
        debug("Returning success payload");
        return Future<dynamic>.value(resBody["payload"]);
      } else if (response.statusCode < 400 && resBody.containsKey("status") && resBody["status"] < 400 ) {
        debug("Returning success entire response");
        return Future<dynamic>.value(resBody);
      } else  {
        debug("processResponse error: ${response.body}");
        return throw( VasatError.fromMap(resBody) );
      }
    }

  }

  ///////////////////////////////////////// Search and Sync Methods ////////////////////////////////////////////////////


  Future<List<T>> searchObject<T extends VasatModel<T>>( VasatFromJsonFactory<T> factory, {Map<String, String>? params, bool updateTimestamp = false} ) {
    // Get the vasat object name from the factory. We simply create an empty object to call the getObjectName().
    var vasatObjectName = factory({}).getObjectName();
    // init params if not passed
    if (params == null) params = {};
    // if we don't pass limit in params use the default value
    if (!params.containsKey("limit")) params["limit"] = config.queryLimit.toString();
    String url = getUrlRoot(vasatObjectName);
    return apiGet(url, params: params, returnPayloadOnly: false).then((res) {
      return processObjects<T>(res, factory, updateTimestamp: updateTimestamp);
    });

  }

  Future<bool> syncFromRemote<T extends VasatModel<T>>(VasatFromJsonFactory<T> factory, {Map<String, String>? params, bool updatesOnly = true}) async {

    // Get the vasat object name from the factory. We simply create an empty object to call the getObjectName().
    var vasatObjectName = factory({}).getObjectName();
    
    return storageProvider.getObjectLastTimestamp( vasatObjectName )
      .then((timestamp) {
        if (updatesOnly && timestamp != null) {
          if (params == null) params = {};
          params!["dateUpdated"] = ">$timestamp";
        }
        return searchObject<T>(factory, params: params, updateTimestamp: true);
      }).then((objList) => storageProvider.saveObjects<T>(objList));
  }

  Future<List<T>> loadFromAssets<T extends VasatModel<T>>(VasatFromJsonFactory<T> factory, BuildContext context) async {
    return DefaultAssetBundle.of(context).loadString("assets/${T.toString()}.json")
        .then((data) => processObjects<T>(jsonDecode(data), factory));
  }


  Future<List<T>> processObjects<T extends VasatModel<T>>(Map<String, dynamic> data, VasatFromJsonFactory<T> factory, {bool updateTimestamp = false}) async {

    // data Map can be the entire  Vasat response or just the Payload part of it so we need to check this
    var finalData = data.containsKey("payload") ? data["payload"] : data;
    debug("processObjects data total -> ${finalData["total"]} offset -> ${finalData["offset"]}");
    
    // Check if we need to update the local value of the last updated timestamp for the object.
    if (updateTimestamp && data.containsKey("time") && data["time"] is int) {
      var timestamp = data["time"] as int;
      var vasatObjectName = factory({}).getObjectName();
      debug("processObjects storing new timestamp -> $timestamp for object -> $vasatObjectName");
      await storageProvider.setObjectLastTimestamp(vasatObjectName, timestamp);
    }

    List<T> res = [];
    if (finalData.containsKey("results") && finalData["results"] is List) {
      List<dynamic> payload = finalData["results"];
      debug("processObjects data result length -> ${payload.length}");
      res = payload.map((item) {
        // Mark object as not dirty as it is fresh from the server
        item["dirty"] = false;
        return factory(item);
      }).toList();
    }
    debug("processObjects for ${T.toString()} finished parsing json data");
    return res;
  }

  void debug(String message, { int? wrapWidth }) {
    if (config.debug) debugPrint("Vasat - " + message, wrapWidth:wrapWidth);
  }
}
